package edu.cs2335.lab7.creature;

import java.io.Serializable;
import java.util.EventListener;

import javax.swing.event.EventListenerList;

import edu.cs2335.lab7.data.CreatureData;
import edu.cs2335.lab7.dwboard.DWBoard;
import edu.cs2335.lab7.dwboard.DWBoardLocation;

/**
 * Represents a 'Creature' in Darwin's World
 *
 * @author Brandyn Roemer Date: Mar 7, 2005 Time: 1:27:29 AM
 */
public class DWCreature implements Serializable {
    /** Constant indicating a north facing creature */
    public static final int NORTH = 0;
    /** Constant indicating a south facing creature */
    public static final int SOUTH = 180;
    /** Constant indicating an east facing creature */
    public static final int EAST = 90;
    /** Constant indicating a west facing creature */
    public static final int WEST = 270;


    /** This is the id of the creature base on index in creature array */
    private int id;
    /** The name for this edu.cs2335.lab7.creature */
    private DWSpecies species;
    /** Creature data */
    private CreatureData data;
    /** The board this edu.cs2335.lab7.creature is on, null if none */
    private DWBoard board;
    /** All the listeners that need to be notified of creature move events */
    private EventListenerList moveListeners = new EventListenerList();
    /** The brains of the creature */
    private Interpreter brain;
    
    /**
     * Creates a new creature with an id and species type
     * @param id The index of this creature
     * @param species the species the creature belong to
     */
    public DWCreature(int id, DWSpecies species) {
        this.id = id;
        this.species = species;
        brain = new Interpreter(species.getScript(), this);
    }

    /**
     * Creates a new creature with an id and species type
     * @param id The index of this creature
     * @param species the species the creature belong to
     * @param board The game board
     */
    public DWCreature(int id, DWSpecies species, DWBoard board) {
        this.id = id;
        this.species = species;
        this.board = board;
        brain = new Interpreter(species.getScript(), this);
    }


    /**
     * The creature will ask it brain what the next thing it should is
     * and then try to do it
     */
    public void doTurn() {
        String cmd = brain.getNextCommand();
        if (cmd.equals("hop")) {
            hop();
        } else if (cmd.equals("infect")) {
            infect();
        } else if (cmd.equals("left")) {
            turnLeft();
        } else if (cmd.equals("right")) {
            turnRight();
        } else { // recived an error somewhere in brain
            System.out.println("err:" + cmd);
        }
    }
    
    
    /**
     * Gets the name of this edu.cs2335.lab7.creature
     * @return The species of this creature
     */
    public DWSpecies getSpecies() {
        return species;
    }

    /**
     * Sets the name of this edu.cs2335.lab7.creature
     * @param species The new species of this creature
     */
    public void setSpecies(DWSpecies species) {
        this.species = species;
    }

    /**
     * Gets the board this edu.cs2335.lab7.creature is on/using
     * @return The board
     */
    public DWBoard getBoard() {
        return board;
    }

    /**
     * Sets the board this edu.cs2335.lab7.creature is on/should use
     *
     * @param board The board
     */
    public void setBoard(DWBoard board) {
        this.board = board;
    }
  
    /**
     * @return Returns the data.
     */
    public CreatureData getData() {
        return data;
    }

    /**
     * @param data The data to set.
     */
    public void setData(CreatureData data) {
        this.data = data;
    }

    /**
     * @return Returns the id.
     */
    public int getId() {
        return id;
    }


    /**
     * Gets the direction this creature is currently facing
     *
     * @return The direction
     */
    public int getDirection() {
        return (data != null ? data.getDirection() : -1);
    }


    /**
     * ajkhsdfilasgf
     * @param row asdfa
     * @param col sdfasdfasdf
     */
    public void setLocation(int row, int col) {

        if (data != null) {

            data.setX(col);
            data.setY(row);
        }
    }


    /**
     * asdf
     * @return asdf
     */
    public int getRow() {
        return (data != null ? data.getY() : -1);
    }


    /**
     * asdf
     * @return asdf
     */
    public int getColumn() {
        return (data != null ? data.getX() : -1);
    }


    /**
     * Adds a CreatureMoveListener to the list to be called
     *
     * @param l The CreatureMoveListener
     */
    public void addCreatureMoveListener(CreatureMoveListener l) {
        moveListeners.add(CreatureMoveListener.class, l);
    }

    /**
     * Removes a CreatureMoveListener from the list to be called
     *
     * @param l The CreatureMoveListener
     */
    public void removeCreatureMoveListener(CreatureMoveListener l) {
        moveListeners.remove(CreatureMoveListener.class, l);
    }

    /**
     * Tells the edu.cs2335.lab7.creature to hop
     *
     * @return True on success, false otherwise
     */
    public boolean hop() {
        boolean moved;
        DWBoardLocation loc;
        CreatureMoveEvent moveEvent;
        loc = getLocationInFront();
        if (loc != null && loc.isOpen()) {
            DWBoardLocation old = board.get(data.getX(), data.getY());
            old.setCreature(null);
            data.setX(loc.getRow());
            data.setY(loc.getCol());
            loc.setCreature(this);
            moved = true;
            moveEvent = new CreatureMoveEvent(CreatureMoveEvent.CREATURE_HOP,
                                              this,
                                              loc);

            notifyCreatureMoveListeners(moveEvent);
        } else {
            moved = false;
        }
        return moved;
    }



    /**
     * Tells the edu.cs2335.lab7.creature to infect
     * @return True on success, false otherwise
     */
    public boolean infect() {

        boolean infected;
        DWBoardLocation loc;
        DWCreature creature;
        CreatureMoveEvent moveEvent;

        if ((loc = getLocationInFront()) != null 
                && (creature = loc.getCreature()) != null) {

            if (!(creature.getSpecies().isequals(this.getSpecies()))) {
                moveEvent = new CreatureMoveEvent(CreatureMoveEvent
                                                  .CREATURE_INFECT,
                              this,
                              creature);
                notifyCreatureMoveListeners(moveEvent);
                creature.infectedBy(this);
                infected = true;
            } else {
             infected = false;   
            }
                
        } else {
            infected = false;
        }
        return infected;
    }



    /**
     * Tells the edu.cs2335.lab7.creature to turn left
     * @return True on success, false otherwise
     */
    public boolean turnLeft() {
        boolean turned;
        CreatureMoveEvent moveEvent;

        switch (data.getDirection()) {
            case (DWCreature.NORTH):
                data.setDirection(DWCreature.WEST);
                turned = true;
                break;
            case (DWCreature.EAST):
                data.setDirection(DWCreature.NORTH);
                turned = true;
                break;
            case (DWCreature.SOUTH):
                data.setDirection(DWCreature.EAST);
                turned = true;
                break;
            case (DWCreature.WEST):
                data.setDirection(DWCreature.SOUTH);
                turned = true;
                break;
            default:
                turned = false;
                break;
        }

        if (turned) {
            moveEvent = new CreatureMoveEvent(CreatureMoveEvent.CREATURE_TURN,
                                              this, data.getDirection(),
                                              CreatureMoveEvent.TURN_LEFT);
            notifyCreatureMoveListeners(moveEvent);
        }
        return turned;
    }



    /**
     * Tells the edu.cs2335.lab7.creature to turn right
     *
     * @return True on success, false otherwise
     */
    public boolean turnRight() {
        boolean turned;
        CreatureMoveEvent moveEvent;

        switch (data.getDirection()) {
        case (DWCreature.NORTH):
            data.setDirection(DWCreature.EAST);
            turned = true;
            break;
        case (DWCreature.EAST):
            data.setDirection(DWCreature.SOUTH);
            turned = true;
            break;
        case (DWCreature.SOUTH):
            data.setDirection(DWCreature.WEST);
            turned = true;
            break;
        case (DWCreature.WEST):
            data.setDirection(DWCreature.NORTH);
            turned = true;
            break;
        default:
            turned = false;
            break;
    }
        if (turned) {
            moveEvent = new CreatureMoveEvent(CreatureMoveEvent.CREATURE_TURN,
                                              this, data.getDirection(),
                                              CreatureMoveEvent.TURN_RIGHT);
            notifyCreatureMoveListeners(moveEvent);
        }
        return turned;
    }


    /**
     * Asks the edu.cs2335.lab7.creature if there's a wall in front of it
     * @return True if there is, false if not
     */
    public boolean wallInFront() {
        DWBoardLocation loc;
        if ((loc = getLocationInFront()) != null) {
            return loc.isWall();
        } else {
            return false;
        }
    }

    /**
     * Asks the edu.cs2335.lab7.creature if there's an enemy in front of it
     * @return True if there is, false if not
     */
    public boolean enemyInFront() {
        DWBoardLocation loc;
        if ((loc = getLocationInFront()) != null) {
            return (loc.getCreature() != null);
        } else {
            return false;
        }
    }



    /**
     * Tells this edu.cs2335.lab7.creature that it was infected by the
     * provided edu.cs2335.lab7.creature
     *
     * @param creature The edu.cs2335.lab7.creature infecting this one
     */
    public void infectedBy(DWCreature creature) {
        CreatureMoveEvent moveEvent;
        if (creature != null) {
            moveEvent = new CreatureMoveEvent(
                                        CreatureMoveEvent.CREATURE_INFECTED_BY,
                                        creature,
                                        this);
            notifyCreatureMoveListeners(moveEvent);
            /* mutate ... */
            setSpecies(creature.getSpecies());
            data.setSpecies(species.getSpeciesType());
            brain.setCCode(species.getScript());
        }

    }

    /**
     * Notifies all the listeners of the CreatureMoveEvent
     * @param e The CreatureMoveEvent to notify the listeners about
     */
    private void notifyCreatureMoveListeners(CreatureMoveEvent e) {
        CreatureMoveListener l;
        EventListener[] lArr;
        lArr = moveListeners.getListeners(CreatureMoveListener.class);
        if (lArr != null) {
            for (int i = 0; i < lArr.length; i++) {
                l = (CreatureMoveListener) lArr[i];
                l.handleCreatureMove(e);
            } // ~for ( [each listener] )
        } // ~if (lArr != null)
    }

    /**
     * Gets the board location directly in front of this creature
     * @return The board location, or null if none exists there
     */
    public DWBoardLocation getLocationInFront() {
        DWBoardLocation loc = null;

        switch (data.getDirection()) {
            case (DWCreature.NORTH):
                if (board != null) {
                    loc = board.get(data.getX(), data.getY() - 1);
                }
                break;
            case (DWCreature.EAST):
                if (board != null) {

                    loc = board.get(data.getX() + 1, data.getY());
                }
                break;
            case (DWCreature.SOUTH):
                if (board != null) {
                    loc = board.get(data.getX(), data.getY() + 1);
                }
                break;
            case (DWCreature.WEST):
                if (board != null) {
                    loc = board.get(data.getX() - 1, data.getY());
                }
                break;
            default:
                loc = null;
                break;
        }
        return loc;
    }
}

