// display an error if user is already playing or watching a game and
// tries to play/watch

package edu.cs2335.lab7.server;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.SocketException;
import java.util.List;

import edu.cs2335.lab7.creature.DWSpecies;
//import edu.cs2335.lab7.data.CreatureData;
import edu.cs2335.lab7.data.MapData;

/**
 * Darwin's World - CS2335 Lab7
 * @version 1.0
 * 
 * ServerConnection - The server connection class. Handles one connection to one
 * client and transfers data to and from them.
 */
public class ServerConnection implements Runnable {
    
    /**Success code.*/
    private static int success = 1;
    
    /**Code to indicate an invalid user was specified by some action. */
    private static int nouser = 2;
    
    /**Code to indicate an invalid password was entered by the user. */
    private static int badpass = 6;
    
    /**Code to indicate a username specified by some action already exists.*/
    private static int duplicateuser = 5;
    
    /**Code to indicate a username is offline.*/
    private static int useroffline = 7;
    
    /**Code to indicate that no such game exists.*/
    private static int nogame = 9;
    
    /**The client socket.*/
    private Socket connection;
    
    /**The main server class.*/
    private ServerRoot server;
    
    /**Input stream to read text from client.*/
    private BufferedReader input;
    
    /**Output stream to write text to client.*/
    private PrintWriter output;
   
    /**The username of the connected user associated with this connection.*/
    private String username;
    
    /**A message received from the client.*/
    private String[] message;
    
    /**A reference to the user database.*/
    private UserDatabase database;

    /**Input stream to read objects from client.*/
    private ObjectInputStream objectInput;
    
    /**Output stream to write objects to server.*/
    private ObjectOutputStream objectOutput;

    /**Name of the game the user is in.*/
    private String gameName;

    /**A boolean that controls whether or not the
     * thread continues at the start of each loop.*/
    private boolean run;
    
    /**
     * Constructor which is called from the main server
     * class when a new connection is accepted.
     * @param connection the client socket
     * @param server and instance of the main server class
     * @param database a refernce to the user database
     * @param firstmessage the first message received from the client
     */
    public ServerConnection(Socket connection, ServerRoot server,
            UserDatabase database, String[] firstmessage) {
        message = firstmessage;
        this.database = database;
        this.connection = connection;
        this.server = server;
        username = "no_one";
    }
    
    /**
     * Main thread loop. Keeps looping and accepting messages until 
     * client quits.
     */
    public void run() {
        run = true;
        try {
        input = new BufferedReader(
                new InputStreamReader(connection.getInputStream()));
        output = new PrintWriter(connection.getOutputStream());
        } catch (IOException e) {
            e.printStackTrace();
        }
        parseMessage(message);
        while (run) {
            try {
            String capturedmessage = input.readLine();
            if (capturedmessage == null) {
                run = false;
                if (username != null) {
                    changeUserStatus("offline");
                }
                continue;
            }
            message = capturedmessage.split("~");
            if (message == null) {
                run = false;
                if (username != null) {
                    changeUserStatus("offline");
                }
                continue;
            }
            parseMessage(message);
            } catch (SocketException se) {
                se.printStackTrace();
                closeConnection();
            } catch (IOException e) {
                e.printStackTrace();
                closeConnection();
            }
        }
    }
    /**
     * Parses the command from the connected user and determines
     * what action to take based on that.
     * @param text the text received from the client 
     */
    private void parseMessage(String[] text) {
        if (text[0].equals("READY")) {
            ready();
        } else if (text[0].equals("LOGIN")) {
            loginToServer(text);
        } else if (text[0].equals("MAINMSG")) {
            sendMainChatMsg(text);
        } else if (text[0].equals("GETUSERSTAT")) {
            getUserStat(text);
        } else if (text[0].equals("GAMEMSG")) {
            sendGameMessage(text);
        } else if (text[0].equals("NEWUSER")) {
            registerUser(text);
        } else if (text[0].equals("GAMEOPTIONS")) {
            setupGame(text);
        } else if (text[0].equals("CHALLENGE")) {
            challengeUser(text);
        } else if (text[0].equals("DECLCHAL")) {
            declineChallenge(text);
        } else if (text[0].equals("ACPTCHAL")) {
            acceptChallenge(text);
        } else if (text[0].equals("JOINGAME")) {
            joinGame(text);
        } else if (text[0].equals("LEAVEGAME")) {
            leaveGame(text);
        } else if (text[0].equals("ADDCREAT")) {
            addCreature(text);
        } else if (text[0].equals("SELCREAT")) {
            selCreature(text);
        } else if (text[0].equals("LISTCREAT")) {
            listCreatures();
        } else if (text[0].equals("DELCREAT")) {
            deleteCreature(text);
        } else if (text[0].equals("QUIT")) {
            closeConnection();
        } else {
            sendMessage("X~UF~unknownerror");
        }
    }

    /**
     * Notifies the server that this client is ready for the next move.
     */
    private void ready() {
        server.setReadyMove(gameName);
        }
  

    /**
     * Gets the statistics of the user.
     * @param command The user's name.
     */
    private void getUserStat(String[] command) {
        // check if player is already playing/watching a game
        for (int i = 0; i < server.getGameList().size(); i++) {
            Game aGame = (Game) server.getGameList().get(i);
            for (int j = 0; j < aGame.getParticipants().size(); j++) {
                if (username.equals(((ServerConnection) aGame.getParticipants()
                                    .get(j)).getUsername())) {
                    sendMessage("CANTCHAL");
                    return;
                }
            }
        }

        // check to see if either player's stables are empty
        List aStable = database.getUser(username).getStable();
        List aStable2 = database.getUser(command[1]).getStable();
        if (aStable.size() == 0 || aStable2.size() == 0) {
            sendMessage("NOCREAT");
            return;
        }

        // locate the user's info
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (command[1].equals(((User) userList.get(i)).getUsername())) {
                sendMessage("USERSTAT~" + command[1] + "~"
                        + ((User) userList.get(i)).getWins() + "~"
                        + ((User) userList.get(i)).getLosses());
            }
        }        
    }

    /**
     * Sends the game information.
     * @param command Number of creatures, map name, and creature name.
     */
    private void setupGame(String[] command) {
        // MapImage map, DWSpecies species1, DWSpecies species2, 
        // int numcreatures, CreatureData[] data
        MapData map = null;
        DWSpecies species2 = null;
        int numcreatures = 0;

        // map data
        map = server.getMap(command[2]);
        if (map == null) {
            System.out.println("map is null");
        }

        // species2
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (username.equals(((User) userList.get(i)).getUsername())) {
                List stable = ((User) userList.get(i)).getStable();
                for (int j = 0; j < stable.size(); j++) {
                    if (command[3].equals(((DWSpecies) stable.get(j))
                            .getSpeciesType())) {
                         species2 = (DWSpecies) stable.get(j);  
                    }
                }
            }
        }
        if (species2 == null) {
            // user wasn't found in the database
            sendMessage("X~UF~unknownerror");
            return;
        }

        // numcreatures
        numcreatures = Integer.parseInt(command[1]);

         
        // add the game options to the game's data
        setupGameData(map, null, species2, numcreatures);
    }

    
    /**
     * Sends game data after it's all collected
     * @param map The map infomation
     * @param species1 The first species
     * @param species2 The second species
     * @param numcreatures The number of creatures on the map
     */
    private synchronized void setupGameData(MapData map, DWSpecies species1,
                               DWSpecies species2, int numcreatures) {
        Game aGame = null;

        // find the game
        for (int n = 0; n < server.getGameList().size(); n++) {
            aGame = (Game) server.getGameList().get(n);
            if (gameName.equals(aGame.getName())) {
                // set info
                if (map != null) {
                    System.out.println("setting map");
                    System.out.println(map.getName());
                     aGame.setMapInfo(map);
                }

                if (species1 != null) {
                    aGame.setSpecies1(species1);
                }
                if (species2 != null) {
                    aGame.setSpecies2(species2);   
                }
                if (numcreatures != 0) {
                    aGame.setNumCreatures(numcreatures);
                }

                // everyone's info has been sent, start shit
                if (aGame.getSpecies1() != null && aGame.getSpecies2() != null
                    && gameName != null) {
                    // initialize the game
                    System.out.println("Initializating the game now.");
                    int gamecode = aGame.initGame();
                    if (gamecode == 1) {
                        System.out.println("Initialization good.");
                    } else {
                        System.out.println(gamecode);
                    }

                    // <- STARTGAME <gamedata>
                    // sending an object, so wait after sending STARTGAME
                    System.out.println("Sending the data now.");
                    for (int m = 0; m < aGame.getParticipants().size();
                            m++) {
                        ((ServerConnection) aGame.getParticipants().get(m))
                                 .sendMessage("STARTGAME");
                    }
                    try {
                        wait(1000);
                    } catch (InterruptedException e) {
                        return;
                    }
                    server.sendGameData(aGame);

                    // update the list of games & update the users in the game
                    try {
                        wait(500);
                    } catch (InterruptedException e) {
                        return;
                    } 
                    server.updateGameLists();
                    server.updateGameUsers(gameName);

                    // @ <- from  GAMEUSERS <user1> <user2> ... <user n> UE
                    String gameUsers = "GAMEUSERS~";
                    for (int m = 0; m < aGame.getParticipants().size();
                            m++) {
                        gameUsers += ((ServerConnection) aGame
                                      .getParticipants().get(m))
                                      .getUsername() + "~";
                    }
                    gameUsers += "UE";
                    for (int m = 0; m < aGame.getParticipants().size();
                            m++) {
                        ((ServerConnection) aGame.getParticipants().get(m))
                                 .sendMessage(gameUsers);
                    }

                    // * <- from ACTIVEGAMES <game1> <game2>...<game n> GE
                    String activeGames = "ACTIVEGAMES~";
                    for (int m = 0; m < server.getGameList().size(); m++) {
                        activeGames += ((Game) server.getGameList().get(m))
                                .getName() + "~";
                    }
                    activeGames += "GE";
                    List userList = database.getUserList();
                    for (int m = 0; m < userList.size(); m++) {
                        String user = ((User) userList.get(m))
                                .getUsername();
                        server.sendMsgToUser(user, activeGames);
                    }
                }
            }
        }
        if (aGame == null) {
            sendMessage("X~UF~unknownerror");   
        }
    }
    
    /**
     * Sending game data to user
     * @param data the data to send
     */
    public void sendData(GameData data) {
        try {
        if (objectOutput == null) {
            objectOutput = new ObjectOutputStream(connection
                    .getOutputStream());
        }
        objectOutput.writeObject(data);
        objectOutput.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Authenticates the client with the user database and logs
     * them in if the information provided is valid. 
     * @param loginstring  the login command received
     */
    private void loginToServer(String[] loginstring) {
        if (loginstring.length != 3) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String uname = loginstring[1];
        String password = loginstring[2];
        if (server.getUserConnection(uname) != null) {
            sendMessage("X~LF~loggedon");
        } else {
            int result = database.loginUser(uname, password);
            if (result == success) {
                sendMessage("LS");
                username = uname;
                changeUserStatus("active");
            } else if (result == nouser) {
                sendMessage("X~LF~unknownuser");
            } else if (result == badpass) {
                sendMessage("X~LF~badpassword");
            }
        }
    }
    
    /**
     * Registers a new username and password in the user database.
     * @param userstring  the registration info received
     */
    private void registerUser(String[] userstring) {
        if (userstring.length != 3) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String uname = userstring[1];
        String password = userstring[2];
        int result = database.registerNewUser(uname, password);
        if (result == success) {
            sendMessage("RS");
        } else if (result == duplicateuser) {
            sendMessage("X~RF~userexists");
        }
    }
    
    /**
     * Changes the user's state and tells the main
     * server class to notify all connected users.
     * @param newstate ???
     */
    private synchronized void changeUserStatus(String newstate) {
        database.changeStatus(username, newstate);
        if (newstate.equals("offline")) {
            server.disconnectUser(username);
        }
        try {
            wait(1000);
        } catch (InterruptedException e) {
            return;
        }
        server.updateMainUsers();
        server.updateGameLists();
    }

    /**
     * Sends a chat message from this client to the main lobby chat
     * window of all connected clients.
     * @param imessage  the message to send
     */
    private void sendMainChatMsg(String[] imessage) {
        if (imessage.length != 2) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String msgtosend = imessage[1];
        server.sendMainMsg("MAINMSG~" + username + "~" + msgtosend);
    }
    
    /**
     * Sends a chat message received from a game chatroom to
     * all other users connected to that game.
     * @param command  message received from client
     */
    private void sendGameMessage(String[] command) {
        if (command.length < 2) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String msg = "GAMEMSG~" + username 
        + "~" + command[1];
        int result = server.sendGameMsg(gameName, msg);
        if (result == nogame) {
            sendMessage("X~MF~nosuchgame");
        }
    }

    /**
     * Sends a challenge to another user from the client associated
     * with this connection.
     * @param command  the command received from the client
     */
    private void challengeUser(String[] command) {
        if (command.length != 2) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String challengeduser = command[1];
        int result = 0;

        // send message with wins/losses
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (username.equals(((User) userList.get(i)).getUsername())) {
                result = server.sendMsgToUser(challengeduser, "CHALLENGE~"
                                  + username + "~"
                                  + ((User) userList.get(i)).getWins() + "~"
                                  + ((User) userList.get(i)).getLosses());
                if (result == success) {
                    return;
                } else if (result == useroffline) {
                    sendMessage("X~MF~useroffline");
                }
            }
        }
    }

    /**
     * Sends a message to another user informing them that
     * this user declined their challenge.
     * @param command  the declining challenge command
     */
    private void declineChallenge(String[] command) {
        if (command.length != 2) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        String challenginguser = command[1];
        server.sendMsgToUser(
                challenginguser, "USERDECL~" + username);
    }

    /**
     * When a user accepts a challenge, this method starts a new
     * game and sends them the options to configure it.
     * @param command  the accepting challenge command
     */
    private void acceptChallenge(String[] command) {
        if (command.length != 3) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        
        String maplist = "maplist~";
        String[] mapnames = server.getMaps();
        for (int i = 0; i < mapnames.length; i++) {
            maplist = maplist + mapnames[i] + "~";
        }
        maplist += "ME~";

        // send the appropriate messages
        String speciesList1 = "";
        String speciesList2 = "";
        List userList = database.getUserList();
        for (int j = 0; j < userList.size(); j++) {
            if (command[1].equals(((User) userList.get(j)).getUsername())) {
                List stable = ((User) userList.get(j)).getStable();
                for (int k = 0; k < stable.size(); k++) {
                    speciesList1 = speciesList1
                                 + ((DWSpecies) stable.get(k)).getSpeciesType()
                                 + "~";
                }
                speciesList1 += "CE~";
            } else if (command[2].equals(((User) userList.get(j))
                                 .getUsername())) {
                List stable = ((User) userList.get(j)).getStable();
                for (int k = 0; k < stable.size(); k++) {
                    speciesList2 = speciesList2
                                 + ((DWSpecies) stable.get(k)).getSpeciesType()
                                 + "~";
                }
                speciesList2 += "CE";
            }
        }
        sendMessage("CHOOSEOPTIONS~" + maplist + speciesList1
                  + "2~5~10~15~20");
        server.sendMsgToUser(command[2], "CHOOSECREAT~" +  speciesList2); 
        

        // begin a new game
        gameName = command[2] + " vs. " + command[1];
        server.getUserConnection(command[2]).setGameName(gameName);
        server.newGame(gameName,
                                      command[2],
                                      command[1]);

        // add game name to the game data
        setupGameData(null, null, null, 0);
    }

    /**
     * Sends the current game data when a user decides to join a game.
     * @param command The name of the game.
     */
    private synchronized void joinGame(String[] command) {
        Game aGame = null;

        // sets their game name
        gameName = command[1];

        // check if player is already playing/watching a game
        for (int i = 0; i < server.getGameList().size(); i++) {
            Game temp = (Game) server.getGameList().get(i);
            for (int j = 0; j < temp.getParticipants().size(); j++) {
                if (username.equals(((ServerConnection) temp.getParticipants()
                                    .get(j)).getUsername())) {
                    sendMessage("CANTJOIN");
                    return;
                }
            }
        }

        // locate the game data
        for (int n = 0; n < server.getGameList().size(); n++) {
            if (command[1].equals(((Game) server.getGameList().get(n))
                    .getName())) {
                aGame = (Game) server.getGameList().get(n);
            }
        }
        if (aGame == null) {
            sendMessage("X~UF~unknownerror");
        }

        // <- STARTGAME <gamedata>
        // sending an object, so wait after sending STARTGAME
        sendMessage("STARTGAME");
        try {
            wait(1000);
        } catch (InterruptedException e) {
            return;
        }
        try {
            if (objectOutput == null) {
            objectOutput = new ObjectOutputStream(connection.getOutputStream());
            }
            objectOutput.writeObject(aGame.getGameData());
            objectOutput.flush();
        } catch (IOException e) {
            return;
        }

        // add the watcher
        aGame.addToParticipants(this);

        // update the users
        try {
            wait(1000);
        } catch (InterruptedException e) {
            return;
        }
        server.updateGameUsers(command[1]);
    }
    
    /**
     * Tells the main server class to remove a player from a connected game
     * and then updates the players display in that game.
     * @param command the game to leave
     */
    private void leaveGame(String[] command) {
        if (command.length < 1) {
            sendMessage("X~UF~unknownerror");
            return;
        }
        int result = server.removeFromGame(gameName, this);
        if (result == success) {
            server.updateGameUsers(gameName);
        }
    }

    /**
     * Adds a creature to the user's list of creatures.
     * @param command The command that contains the creature's name.
     */
    private void addCreature(String[] command) {
        String creatureName = command[1];
        DWSpecies newSpecies = null;
        List stable = null;

        // check if the creature exists
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            List userStable = ((User) userList.get(i)).getStable();
            for (int j = 0; j < userStable.size(); j++) {
                if (creatureName.equals(((DWSpecies) userStable.get(j))
                        .getSpeciesType())) {
                    // failure
                    sendMessage("X~CF~creatureexists");
                    return;
                }
            }
        }

        // get the user's stable
        User user = database.getUser(username);
        stable = user.getStable();

        // success
        sendMessage("CS");

        // wait to get creature object
        try {
            if (objectInput == null) {
            objectInput = new ObjectInputStream(connection.getInputStream());
            }
            newSpecies = (DWSpecies) objectInput.readObject();
        } catch (IOException e) {
            return;
        } catch (ClassNotFoundException e) {
            return;
        }

        // add it to the user's stable
        stable.add(newSpecies);

        // success
        listCreatures();
    }

    /**
     * Lists the creatures that the user has.
     */
    private void listCreatures() {
        String newMessage = "CREATLIST~";
        List stable = null;

        // locate the user's stable
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (username.equals(((User) userList.get(i)).getUsername())) {
                stable = ((User) userList.get(i)).getStable();
            }
        }
        if (stable == null) {
            // user wasn't found in the database
            sendMessage("X~UF~unknownerror");
            return;
        }

        // traverse through the user's stable and setup/send the message
        for (int j = 0; j < stable.size(); j++) {
            newMessage = newMessage
                    + ((DWSpecies) stable.get(j)).getSpeciesType()
                    + "  Wins: " + ((DWSpecies) stable.get(j)).getWins()
                    + " Losses: " + ((DWSpecies) stable.get(j)).getLosses()
                    + "~";
        }
        newMessage += "CE";
        sendMessage(newMessage);
    }

    /**
     * Removes a creature from the user's stable.
     * @param command The command that contains the name of the creature.
     */
    private void deleteCreature(String[] command) {
        String creatureName = command[1];
        List stable = null;

        // locate the user's stable
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (username.equals(((User) userList.get(i)).getUsername())) {
                stable = ((User) userList.get(i)).getStable();
            }
        }
        if (stable == null) {
            // user wasn't found in the database
            sendMessage("X~UF~unknownerror");
            return;
        }

        // find the creature and remove it
        for (int j = 0; j < stable.size(); j++) {
            if (creatureName.equals(((DWSpecies) stable.get(j))
                                    .getSpeciesType())) {
                stable.remove(j);
                return;
            }
        }

        // creature wasn't found
        listCreatures();
    }
    
    /**
     * Ends the game--Removes the game from the game list.
     */
    public void endGame() {
        server.endGame(gameName);
    }

    /**
     * Returns the creature object of the name passed in.
     * @param command The name of the creature.
     */
    private void selCreature(String[] command) {
        DWSpecies species = null;
        
        // locate the species
        List userList = database.getUserList();
        for (int i = 0; i < userList.size(); i++) {
            if (username.equals(((User) userList.get(i)).getUsername())) {
                List stable = ((User) userList.get(i)).getStable();
                for (int j = 0; j < stable.size(); j++) {
                    if (command[1].equals(((DWSpecies) stable.get(j))
                            .getSpeciesType())) {
                         species = (DWSpecies) stable.get(j);  
                    }
                }
            }
        }
        if (species == null) {
            // user wasn't found in the database
            sendMessage("X~UF~unknownerror");
            return;
        }

        // add the species to the game's data
        setupGameData(null, species, null, 0);
    }

    /**
     *When a command is receieved to shutdown, this method closes
     *all sockets, changes the user's status to offline, and then sets run
     *to false to end thread execution.
     */
    public void closeConnection() {
        changeUserStatus("offline");
        server.exitGames(username);
        try {
            run = false;
            input.close();
            output.close();
            connection.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Sends a message to the user connected on this client socket.
     * @param messagetosend  the message
     */
    public void sendMessage(String messagetosend) {
        System.out.println(messagetosend);
        output.println(messagetosend);
        output.flush();
    }
    
    /**
     * Returns the username of the player connected to this socket.
     * @return username  the name
     */
    public String getUsername() {
        return username;
    }

    /**
     * @param gameName The gameName to set.
     */
    public void setGameName(String gameName) {
        this.gameName = gameName;
    }
    /**
     * @return Returns the database.
     */
    public UserDatabase getDatabase() {
        return database;
    }
}