package edu.cs2335.lab7.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;

import edu.cs2335.lab7.client.DataUnzip;
import edu.cs2335.lab7.data.MapData;
import edu.cs2335.lab7.data.MapMaker;

/**
 * Darwin's World - CS2335 Lab7
 * @version 1.0
 * 
 * ServerRoot - The main server class. Controls server startup and accepts
 * client connections. Performs different operations involving multiple clients.
 */
public class ServerRoot {
    
    /**Success code.*/
    private static int success = 1;
    
    /**Code to indicate a username specified by some action already exists.*/
    private static int duplicateuser = 5;
    
    /**Code to indicate a username is offline.*/
    private static int useroffline = 7;
    
    /**Code to indicate that no such game exists.*/
    private static int nogame = 9;
    
    /**Main server socket, only one created.*/
    private ServerSocket mainsocket;
    
    /**Client socket, one created for each connected client.*/
    private Socket clientsocket;
    
    /**Buffer to read input from client sockets.*/
    private BufferedReader input;
    
    /**A list of all client connections.*/
    private List connections;
    
    /**Reference to the database of user information.*/
    private UserDatabase database;

    /**A list of all active games.*/
    private List gamelist;

    /**A list of server maps.*/
    private String[] maps = {"bin/map/Basic-20x20.zip",
            "bin/map/Pacman-50x50.zip", "bin/map/TwoBridges-15x15.zip",
            "bin/map/City-15x15.zip", "bin/map/Basic-2x2.zip"};
    
    /**A list of server map files.*/
    private String[] mapnames = {"Basic-20x20", "Pacman-50x50",
            "TwoBridges-15x15", "City-15x15", "Basic-2x2"};
    
    /**
     * Binds server to port given and attempts to read the database
     * from disk. If no database exists it creates a new one.
     * @param port the server port number
     */
    public ServerRoot(int port) {
        
        connections = new ArrayList();
        gamelist = new ArrayList();
        try {
            mainsocket = new ServerSocket(port);
            FileInputStream fileinput = new FileInputStream("userdatabase.db");
            ObjectInputStream objinput = new ObjectInputStream(fileinput);
            database = (UserDatabase) (objinput.readObject());
            objinput.close();
            fileinput.close();
        } catch (FileNotFoundException fe) {
            database = new UserDatabase();
        } catch (ClassNotFoundException ce) {
            ce.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }

    /**
     * Loads all the map located in the map file
     *
     */
    /**
     * @param mapname sdf
     * @return r re
     */
    public MapData getMap(String mapname) {
        File mapfile = null;
        MapData map = null;
        for (int i = 0; i < mapnames.length; i++) {
            if (mapnames[i].equals(mapname)) {
                mapfile = new File(maps[i]);
                
                DataUnzip.unzip(mapfile, mapnames[i]);
                MapMaker newMap = new MapMaker(mapnames[i]);
                int err;
                if ((err = newMap.readInMap()) != 1) {
                    System.out.println("invalid map file" + err);
                    return null;
                }
                map = newMap.getMapInfo();
                return map;
            }
        }
        return null;
    }
    
    /**
     * Begins accepting connections. When a connection is received a new
     * thread is created and the loop repeats to accept more. The only commands
     * that will be accepted here are EXIT for shutting down the server, NEWUSER
     * for someone registering a new username/password, and LOGIN for someone
     * logging in using an existing account. Anything else is ignored.
     */
    public void startServer() {
        try {
            while (true) {
                clientsocket = mainsocket.accept();
                input = new BufferedReader(
                        new InputStreamReader(clientsocket.getInputStream()));
                String readinput = input.readLine();
                if (readinput == null) {
                    continue;
                }
                String[] firstmessage = readinput.split("~");
                if (firstmessage[0].equals("EXIT")) {
                    shutdownServer();
                    return;
                } else if (firstmessage[0].equals("NEWUSER")
                        || firstmessage[0].equals("LOGIN")) {
                    ServerConnection newconnection = 
                        new ServerConnection(
                                clientsocket, this, database, firstmessage);
                    connections.add(newconnection);
                    Thread connectionthread = new Thread(newconnection);
                    connectionthread.start();
                }
            }
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }
    
    /**
     * Sending game data to all players of a game
     * @param agame The game 
     */
    public synchronized void sendGameData(Game agame) {
        ServerConnection aplayer = null;
        List participants = agame.getParticipants();
        for (int i = 0; i < participants.size(); i++) {
            aplayer = (ServerConnection) (participants.get(i));
            aplayer.sendData(agame.getGameData());
        }
    }
    
    /**
     * When the server gets a shutdown message the database is saved to the
     * disk, all connected clients are notified of the shutdown, 
     * and the server exits.
     */
    public void shutdownServer() {
        database.logoffAllUsers();
        ServerConnection cn;
        for (int i = 0; i < connections.size(); i++) {
            cn = (ServerConnection) (connections.get(i));
            cn.sendMessage("SF~SHUTDOWN");
        }
        try {
            FileOutputStream fileoutput = 
                new FileOutputStream("userdatabase.db");
            ObjectOutputStream objoutput = new ObjectOutputStream(fileoutput);
            objoutput.writeObject(database);
            objoutput.flush();
            objoutput.close();
            fileoutput.close();
            mainsocket.close();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }
    
    /**
     * when a client thread detects that a client has closed their connection
     * this method is called to remove them from the connection list
     * @param username the user to disconnect
     */
    public void disconnectUser(String username) {
        ServerConnection conn = null;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(username)) {
                connections.remove(i);
            }
        }
    }
    
    /**
     * @param messagetosend a message
     * sends a message to all connected client sockets
     */
    public void echoMessage(String messagetosend) {
        ServerConnection conn;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            conn.sendMessage(messagetosend);
        }
    }
    
    /**
     * sends a message to one specific client socket
     * @param destuser the user to send the message to
     * @param msgtosend the message to send
     * @return success  the player was located and the message sent
     *         nouser no user with that username could be found in the database
     *         useroffline a user by that name exists but they are offline
     */
    public int sendMsgToUser(String destuser, String msgtosend) {
        ServerConnection conn;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(destuser)) {
                conn.sendMessage(msgtosend);
                return success;
            }
        }
        return useroffline;
    }
    
    /**
     * Gets the user server connection from the user list
     * @param usernm the user name
     * @return there server connection
     */
    public ServerConnection getUserConnection(String usernm) {
        ServerConnection conn;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(usernm)) {
                return conn;
            }
        }
        return null;
    }
    
    /**
     * Creates a new game and adds the starting two players to it.
     * @param gamename the name of this game
     * @param user1 the first user
     * @param user2 the second user
     * @return success game created successfully
     *         duplicategame  a game by that name already exists
     *         useroffline  a specified user is not online
     */
    public synchronized Game newGame(String gamename, String user1, 
            String user2) {
        Game currgame = null;
        ServerConnection player1 = null;
        ServerConnection player2 = null;
        ServerConnection conn = null;
        for (int i = 0; i < gamelist.size(); i++) {
            if ((((Game) (gamelist.get(i)))
                    .getName()).equals(gamename)) {
                return null;
                //return duplicategame;
            }
        }
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(user1)) {
                player1 = conn;
            } else if ((conn.getUsername()).equals(user2)) {
                player2 = conn;
            }
        }
        if (user1 == null || user2 == null) {
            return null;
            //return useroffline;
        }
        currgame = new Game(gamename);
        currgame.addToParticipants(player1);
        currgame.addToParticipants(player2);
        currgame.setServerGame(true);
        gamelist.add(currgame);
        return currgame;
    }
    
    /**
     * Adds a new participant to a game. This is used when other users
     * connect in the middle of a game to watch it.
     * @param gamename  the game the user will be added to
     * @param conn  the server connection of the player being added
     * @return success  player was added successfully
     *         duplicateuser  that player is already connected to this game
     *         nogame  no game with the given name could be found
     */
    public synchronized int addToGame(String gamename, ServerConnection conn) {
        Game destgame = null;
        List gameusers = null;
        for (int i = 0; i < gamelist.size(); i++) {
            if ((((Game) (gamelist.get(i))).getName()).equals(gamename)) {
                destgame = (Game) (gamelist.get(i));
                gameusers = destgame.getParticipants();
            }
        }
        if (gameusers == null) {
            return nogame;
        }
        for (int i = 0; i < gameusers.size(); i++) {
            if ((((ServerConnection) (gameusers.get(i))).getUsername())
                    .equals(conn.getUsername())) {
                return duplicateuser;
            }
        }
        destgame.addToParticipants(conn);
        return success;
    }
    
    /**
     * Removes a player from a game. Occurs whenever a client
     * is connected to a game when they disconnect.
     * @param gamename game to remove player from
     * @param conn the server connection of the player to remove
     * @return success  player removed successfully
     *         nogame  no game with the given name could be found
     */
    public synchronized int removeFromGame(String gamename, 
            ServerConnection conn) {
        Game currgame = null;
        for (int i = 0; i < gamelist.size(); i++) {
            if ((((Game) (gamelist.get(i))).getName()).equals(gamename)) {
                currgame = (Game) (gamelist.get(i));
                currgame.removeFromParticipants(conn);
                if ((currgame.getParticipants()).size() == 0) {
                    gamelist.remove(i);
                    updateGameLists();
                }
                return success;
            }
        }
        return nogame;
    }
    
    /**
     * Updates the list of connected users to a game by broadcasting a new
     * list to all clients connected to that game. The clients then 
     * redisplay the list.
     * @param gamename  the game to update
     */
    public void updateGameUsers(String gamename) {
        String updatemessage = "GAMEUSERS~";
        Game gametoupdate = null;
        List gameusers = null;
        ServerConnection userconnection = null;
        for (int i = 0; i < gamelist.size(); i++) {
            if ((((Game) (gamelist.get(i)))
                    .getName()).equals(gamename)) {
                gametoupdate = (Game) (gamelist.get(i));
            }
        }
        if (gametoupdate == null) {
            return;
        }
        gameusers = gametoupdate.getParticipants();
        if (gameusers.size() == 0) {
            endGame(gamename);
            return;
        }
        for (int i = 0; i < gameusers.size(); i++) {
            updatemessage +=
                ((ServerConnection) (gameusers.get(i))).getUsername() + "~";
        }
        updatemessage += "UE";
        for (int i = 0; i < gameusers.size(); i++) {
            userconnection = (ServerConnection) (gameusers.get(i));
            userconnection.sendMessage(updatemessage);
        }
    }
    
    /**
     * Updates the main list of users in the lobby window.
     * Called when someone logs in or exits.
     */
    public void updateMainUsers() {
        String updatemessage = "ACTIVEUSERS";
        ServerConnection conn = null;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            updatemessage += "~" + conn.getUsername();
        }
        updatemessage += "~UE";
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            conn.sendMessage(updatemessage);
        }
    }
    
    /**
     * Updates the list of games in the lobby window.
     * Called when someone creates a new game.
     */
    public void updateGameLists() {
        String updatemessage = "ACTIVEGAMES";
        for (int i = 0; i < gamelist.size(); i++) {
            updatemessage += "~" + ((Game) gamelist.get(i)).getName();
        }
        updatemessage += "~GE";
        ServerConnection conn = null;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) connections.get(i);
            conn.sendMessage(updatemessage);
        }
    }
    
   /**
    * Sends a chat message to a game chatroom. Communicates a string
    * message to all client sockets that are logged in.
    * @param msg  message to send
    */
    public void sendMainMsg(String msg) {
        ServerConnection conn = null;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            conn.sendMessage(msg);
        }
    }
    
    /**
     * Removes the game and updates the list of games.
     * @param gamename Name of the game being removed.
     */
    public void endGame(String gamename) {
        Game test = null;
        for (int i = 0; i < gamelist.size(); i++) {
            test = (Game) (gamelist.get(i));
            if ((test.getName()).equals(gamename)) {
                gamelist.remove(i);
                updateGameLists();
            }
        }
    }
    
    /**
     * Sends a chat message to a game chatroom. Communicates a string
     * message to all client sockets connected to that game.
     * @param gamename  the name of the game to send the message to
     * @param message the message to send
     * @return success message sent successfully
     *         nogame  game has gone offline or there is no game with given name
     */
    public int sendGameMsg(String gamename, String message) {
        List gameusers = null;
        ServerConnection userconnection = null;
        for (int i = 0; i < gamelist.size(); i++) {
            if ((((Game) (gamelist.get(i)))
                    .getName()).equals(gamename)) {
                gameusers = ((Game) (gamelist.get(i))).getParticipants();
            }
        }
        if (gameusers == null) {
            return nogame;
        }
        for (int i = 0; i < gameusers.size(); i++) {
            userconnection = (ServerConnection) (gameusers.get(i));
            userconnection.sendMessage(message);
        }
        return success;
    }
    
    /**
     * Removes a user from any connected game when they quit.
     * @param username the exiting user
     */
    public void exitGames(String username) {
        ServerConnection conn = null;
        Game game = null;
        List gameusers = null;
        for (int i = 0; i < gamelist.size(); i++) {
            game = (Game) (gamelist.get(i));
            gameusers = game.getParticipants();
            for (int j = 0; j < gameusers.size(); j++) {
                conn = (ServerConnection) (gameusers.get(j));
                if ((conn.getUsername()).equals(username)) {
                    game.removeFromParticipants(conn);
                    updateGameUsers(game.getName());
                }
            }
        }
    }
    
    /**
     * Tells the game that we are ready for the next move
     * @param gamename the game to talk to
     */
    public void setReadyMove(String gamename) {
        Game testgame = null;
        for (int i = 0; i < gamelist.size(); i++) {
            testgame = ((Game) (gamelist.get(i)));
            if (gamename.equals(testgame.getName())) {
                testgame.addReadyUser();
            }
        }
    }
    
    /**
     * Starts the server from the command line.
     * Prints a usage error if port number is entered incorrectly.
     * @param args port number
     */
    public static void main(String[] args) {
        ServerRoot server;
        int port = 0;
        if (args.length != 1) {
            System.out.println("Usage: <portnumber>");
            System.exit(1);
        } else {
            try {
                String portnum = args[0];
                port = Integer.parseInt(portnum);
                server = new ServerRoot(port);
                server.startServer();
            } catch (NumberFormatException e) {
                System.out.println("You did not enter a valid port.");
                System.exit(1);
            }
        }
    }
    
    /**
     * @return The list of games.
     */
    public List getGameList() {
        return gamelist;
    }
    
    /**
     * Gets a list of maps
     * @return the map list
     */
    public String[] getMaps() {
        return mapnames;
    }
}